import pandas as pd


class KX:
    @staticmethod
    def const(kx:float, E_mod:float=1) -> float:
        return E_mod * kx

    @staticmethod
    def constRectangular(dx_glob:float, dy_glob:float, E_mod:float=1) -> float:
        return E_mod * dy_glob * dx_glob**3 / 12 #EIy
    
    @staticmethod
    def from_csv(csv_path:str, momYColName:str, EIYColName:str) -> pd.DataFrame:
        df_all = pd.read_csv(csv_path)
        df_momEIy = df_all[[momYColName, EIYColName]]
        return df_momEIy
    
    
class KY:
    @staticmethod
    def const(kx:float) -> float:
        return kx

    @staticmethod
    def constRectangular(dx_glob:float, dy_glob:float, E_mod:float=1) -> float:
        return E_mod * dx_glob * dy_glob**3 / 12 #EIx
    
    @staticmethod
    def from_csv(csv_path:str, momXColName:str, EIXColName:str) -> pd.DataFrame:
        df_all = pd.read_csv(csv_path)
        df_momEIy = df_all[[momXColName, EIXColName]]
        return df_momEIy


import pandas as pd
import numpy as np

from .node import SupportNode
from .structure import Stucture
from .utils import interpolateXY
from .linsolve import LinSolve

class NonLinSolve:
    """
    A class for non-linear solving of structural problems.

    This class implements a non-linear solver for structural analysis, 
    considering the effects of geometric non-linearity.

    Parameters
    ----------
    structure : Stucture
        The structural model to be analyzed.
    x_mass_force : float, optional
        The mass force in the x-direction (default is 1).
    y_mass_force : float, optional
        The mass force in the y-direction (default is 1).
    iterations : int, optional
        The number of iterations for the non-linear solution (default is 40).
    z_heigt : float, optional
        The height in the z-direction for moment calculations (default is 1).
    verbose : bool, optional
        If True, print iteration progress (default is True).

    Attributes
    ----------
    _structure : Stucture
        The structural model being analyzed.
    _x_force : float
        The mass force in the x-direction.
    _y_force : float
        The mass force in the y-direction.
    _iterations : int
        The number of iterations for the non-linear solution.
    _z_heigt : float
        The height in the z-direction for moment calculations.
    _verbose : bool
        Flag for verbose output.
    _node_tracker : dict
        Dictionary to track node properties across iterations.
    _structure_tracker : dict
        Dictionary to track structure properties across iterations.
    _table : pd.DataFrame
        DataFrame containing all tracked data across iterations.
    _table_onlyUpdates : pd.DataFrame
        DataFrame containing only the data that changed across iterations.
    """      
    def __init__(
            self,
            structure:Stucture,
            x_mass_force:float=1,
            y_mass_force:float=1,
            iterations:int=20,
            z_heigt:float=1,
            verbose:bool=True
            ) -> None:
        self._structure = structure
        self._x_force = x_mass_force
        self._y_force = y_mass_force
        self._iterations = iterations
        self._z_heigt = z_heigt

        self._verbose = verbose
        
        self._main()


    def _init_node_tracker(self) -> None:
        self._node_tracker = {}
        for node in self._structure._linnodes:
            self._update_node_tracker(node, init=True)


    def _append_node_tracker(self) -> None:
        for node in self._structure._linnodes:
            self._update_node_tracker(node)


    def _update_node_tracker(self, node:SupportNode, init=False) -> None:
        TRACKING_REGISTER = {
            f'node {node._nr} EIx':[node._glob_EIx],
            f'node {node._nr} EIy':[node._glob_EIy],
            f'node {node._nr} Vx':[-node._Rx],
            f'node {node._nr} Vy':[-node._Ry],
            f'node {node._nr} Mx':[-node._Ry * self._z_heigt],
            f'node {node._nr} My':[-node._Rx * self._z_heigt]
        }
        if init:
            for key, item in TRACKING_REGISTER.items():
                self._node_tracker[key] = item
        else:
            for key, item in TRACKING_REGISTER.items():
                self._node_tracker[key].append(item[0])



    def _init_structure_tracker(self) -> None:
        self._structure_tracker = {}
        self._update_structure_tracker(self._structure, init=True)


    def _append_structure_tracker(self) -> None:
        self._update_structure_tracker(self._structure)


    def _update_structure_tracker(self, structure:Stucture, init=False) -> None:
        TRACKING_REGISTER = {
            f'x_s':[structure._loc_stiff_centre_x],
            f'y_s':[structure._loc_stiff_centre_y],
        }
        if init:
            for key, item in TRACKING_REGISTER.items():
                self._structure_tracker[key] = item
        else:
            for key, item in TRACKING_REGISTER.items():
                self._structure_tracker[key].append(item[0])


    
        
    def _update_linnodes_inplace(self) -> None:
        nodes_and_linnodes = zip(
            self._structure._nodes,
            self._structure._linnodes
            )
        for node, linnode in nodes_and_linnodes:
            if isinstance(node._glob_EIx, pd.DataFrame):
                linnode._glob_EIx = interpolateXY(
                    node._glob_EIx,
                    -linnode._Ry*self._z_heigt
                    )
            if isinstance(node._glob_EIy, pd.DataFrame):
                linnode._glob_EIy = interpolateXY(
                    node._glob_EIy,
                    -linnode._Rx*self._z_heigt
                    )


    def _linsolve_inplace(self) -> None:
        sol = LinSolve(self._structure, self._x_force, self._y_force)
        sol.updateNodes()


    def _iterate(self) -> None:
        self._init_structure_tracker()
        self._init_node_tracker()
        for i, _ in enumerate(range(self._iterations)):
            if self._verbose:
                print(f"-> iteration {i+1}/{self._iterations}", end='\r')
            self._update_linnodes_inplace()
            self._linsolve_inplace()
            self._append_structure_tracker()
            self._append_node_tracker()


    def _build_tracking_df(self) -> pd.DataFrame:
        stru_df = pd.DataFrame(self._structure_tracker)
        node_df = pd.DataFrame(self._node_tracker)
        return pd.concat([stru_df, node_df], axis=1)

    
    def _update_nodes_only(self, table:pd.DataFrame) -> pd.DataFrame:
        table_onlyUpdates = table.loc[
            :, table.iloc[0, :] != table.iloc[-1, :]
        ]
        return table_onlyUpdates
    

    def printStructureTable(self) -> None:
        """
        Print the structure table.

        This method calls the printTable method of the structure object
        to display its current state.

        Returns
        -------
        None
        """
        self._structure.printTable()

    
    def printResultTable(self) -> None:
        """
        Print the result table after linear solving.

        This method creates a new LinSolve object with the current structure
        and forces, then prints its result table.

        Returns
        -------
        None
        """
        sol = LinSolve(
            self._structure,
            self._x_force,
            self._y_force
        )
        sol.printTable()


    def printIterationTable(self) -> None:
        """
        Print the table of updates across iterations.

        This method prints the _table_onlyUpdates attribute, which contains
        only the data that changed across iterations.

        Returns
        -------
        None
        """
        print(self._table_onlyUpdates)


    def _main(self) -> None:

        self._linsolve_inplace()
        self._iterate()

        self._table = self._build_tracking_df()
        self._table_onlyUpdates = self._update_nodes_only(self._table)
        
    





import pandas as pd
import numpy as np
from copy import deepcopy

from .polygon import Polygon
from .stiffnesses import KX, KY
from .node import SupportNode
from .utils import interpolateXY

class Stucture:
    """
    A class to represent a structural system defined by a polygon and a list of support nodes.

    Parameters
    ----------
    polygon : Polygon
        The polygon that defines the structural geometry.
    nodes : list of SupportNode
        A list of support nodes associated with the structure.
    
    Attributes
    ----------
    _polygon : Polygon
        The polygon object representing the structural geometry.
    _nodes : list of SupportNode
        List of support nodes forming the structure.
    _node_numbers : pd.Series
        Series representing the node numbers of the structure.
    _node_x : pd.Series
        Series representing the x-coordinates of the nodes.
    _node_y : pd.Series
        Series representing the y-coordinates of the nodes.
    _node_EIy : pd.Series
        Series representing the flexural stiffness of nodes along the y-axis.
    _node_EIx : pd.Series
        Series representing the flexural stiffness of nodes along the x-axis.
    _node_diff_x_xm : pd.Series
        Series representing the difference between node x-coordinates and the
        polygon centroid x-coordinate.
    _node_diff_y_ym : pd.Series
        Series representing the difference between node y-coordinates and the
        polygon centroid y-coordinate.
    _stiff_centre_x : float
        The x-coordinate of the stiffness center of the structure.
    _stiff_centre_y : float
        The y-coordinate of the stiffness center of the structure.
    _node_diff_xs_xm : pd.Series
        Series representing the difference between the stiffness center
        x-coordinate and node x-coordinates.
    _node_diff_ys_ym : pd.Series
        Series representing the difference between the stiffness center
        y-coordinate and node y-coordinates.
    _node_EIx_proportion : pd.Series
        Series representing the proportion of the total flexural stiffness
        along the x-axis for each node.
    _node_EIy_proportion : pd.Series
        Series representing the proportion of the total flexural stiffness
        along the y-axis for each node.
    _global_EIw : pd.Series
        Series representing the global warping stiffness.
    _node_EIwx_proportion : pd.Series
        Series representing the proportion of the torsional stiffness
        contribution for each node along the x-axis.
    _node_EIwy_proportion : pd.Series
        Series representing the proportion of the torsional stiffness
        contribution for each node along the y-axis.
    _result_table : pd.DataFrame
        DataFrame containing various structural properties and node data.
    """  
    def __init__(
            self,
            nodes:list[SupportNode],
            glo_mass_centre:tuple[float, float]|np.ndarray,
            verbose:bool=True
            ):
        
        self._nodes = nodes
        self._glo_mass_centre_x, self._glo_mass_centre_y = glo_mass_centre
        self._verbose = verbose
        self._linnodes = self._to_linear_nodes(deepcopy(nodes))


    def _to_linear_nodes(self, nodes:list[SupportNode]) -> list[SupportNode]:

        MOMENTUM = 0

        def printInfo(nr:int, axis:str, EI:float) -> None:
            if self._verbose:
                print(
                    f"Info: node {nr} -> took EI{axis}(Mom={MOMENTUM}) "
                    f"= {EI:,.1f} for linear solving"
                    )

        def extractStiffnessAtMomentZero(node:SupportNode) -> SupportNode:
            if isinstance(node._glob_EIx, pd.DataFrame):
                node._glob_EIx = interpolateXY(node._glob_EIx, MOMENTUM)
                printInfo(node._nr, 'x', node._glob_EIx)
            if isinstance(node._glob_EIy, pd.DataFrame):
                node._glob_EIy = interpolateXY(node._glob_EIy, MOMENTUM)
                printInfo(node._nr, 'y', node._glob_EIy)
            return node
        
        return [extractStiffnessAtMomentZero(node) for node in nodes]
    

    @property
    def _node_numbers(self) -> pd.Series:
        return pd.Series([node._nr for node in self._linnodes])
    
    @property
    def _glo_node_x(self) -> pd.Series:
        return pd.Series([node._glob_x for node in self._linnodes])

    @property
    def _glo_node_y(self) -> pd.Series:
        return pd.Series([node._glob_y for node in self._linnodes])
    
    @property
    def _loc_node_x(self) -> pd.Series:
        return self._glo_node_x - self._glo_mass_centre_x
    
    @property
    def _loc_node_y(self) -> pd.Series:
        return self._glo_node_y - self._glo_mass_centre_y

    @property
    def _node_EIy(self) -> pd.Series:
        return pd.Series([node._glob_EIy for node in self._linnodes])
    
    @property
    def _node_EIx(self) -> pd.Series:
        return pd.Series([node._glob_EIx for node in self._linnodes])
    
    @property
    def _loc_stiff_centre_x(self) -> float:
        x_xm = self._loc_node_x
        EIx = self._node_EIx
        return (EIx * x_xm).sum() / EIx.sum()

    @property
    def _loc_stiff_centre_y(self) -> float:
        y_ym = self._loc_node_y
        EIy = self._node_EIy
        return (EIy * y_ym).sum() / EIy.sum()

    @property
    def _glo_stiff_centre_x(self) -> float:
        return self._loc_stiff_centre_x + self._glo_mass_centre_x

    @property
    def _glo_stiff_centre_y(self) -> float:
        return self._loc_stiff_centre_y + self._glo_mass_centre_y

    @property
    def _loc_node_xs(self) -> pd.Series:
        return self._glo_node_x - self._glo_stiff_centre_x
    
    @property
    def _loc_node_ys(self) -> pd.Series:
        return self._glo_node_y - self._glo_stiff_centre_y
    
    @property
    def _node_EIx_proportion(self) -> pd.Series:
        return self._node_EIy / self._node_EIy.sum()
    
    @property
    def _node_EIy_proportion(self) -> pd.Series:
        return self._node_EIx / self._node_EIx.sum()
    
    @property
    def _global_EIw(self) -> pd.Series:
        EIy = self._node_EIy
        EIx = self._node_EIx
        x = self._loc_node_xs
        y = self._loc_node_ys
        return (EIy*y**2 + EIx*x**2).sum()
    
    @property
    def _node_EIwx_proportion(self) -> pd.Series:
        return self._node_EIy * self._loc_node_ys / self._global_EIw
    
    @property
    def _node_EIwy_proportion(self) -> pd.Series:
        return self._node_EIx * self._loc_node_xs / self._global_EIw
    
    @property
    def _result_table(self) -> pd.DataFrame:

        result_table = {
            'node nr':self._node_numbers,
            'glo x':self._glo_node_x,
            'glo y':self._glo_node_y,
            'loc x':self._loc_node_x,
            'loc y':self._loc_node_y,
            'loc xs ':self._loc_node_xs,
            'loc ys':self._loc_node_ys,
            'EIx':self._node_EIx,
            'EIy':self._node_EIy,
            '% EIx':self._node_EIx_proportion,
            '% EIy':self._node_EIy_proportion,
            '% EIwx':self._node_EIwx_proportion,
            '% EIwy':self._node_EIwy_proportion
        }

        return pd.DataFrame(result_table)


    def printTable(self) -> None:
        """
        Prints a detailed summary of the structure's geometric and stiffness properties.

        This method outputs the polygon area, centroid, stiffness center, total stiffness 
        in x and y directions, as well as the global warping stiffness. It also prints 
        a table with node-specific data such as coordinates, stiffness proportions, 
        and torsional stiffness contributions.

        Returns
        -------
        None
        """
        print(
            "\n"
            f"glo mass   centre [x,y] : "
            f"{self._glo_mass_centre_x:0.4f}, "
            f"{self._glo_mass_centre_y:0.4f}\n"
            f"glo stiff. centre [x,y] : "
            f"{self._glo_stiff_centre_x:0.4f}, "
            f"{self._glo_stiff_centre_y:0.4f}\n"
            f"loc stiff. centre [x,y] : "
            f"{self._loc_stiff_centre_x:0.4f}, "
            f"{self._loc_stiff_centre_y:0.4f}\n"
            f"EIx total               : "
            f"{self._node_EIx.sum():,.1f}\n"
            f"EIy total               : "
            f"{self._node_EIy.sum():,.1f}\n"
            f"EIw total               : "
            f"{self._global_EIw:,.1f}\n"
            f"\n{self._result_table}\n"
            )


from scipy.interpolate import interp1d
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.axes as mpl_axes
import matplotlib.figure as mpl_fig
from datetime import datetime


def interpolateXY(df:pd.DataFrame, Momentum:float|int) -> float:
    
    x_val = df['mom']
    y_val = df['EI']

    interpolation_function = interp1d(
        x_val,
        y_val,
        kind='linear',
        fill_value="extrapolate"
        )

    stiffness:np.ndarray = interpolation_function(Momentum)

    return float(stiffness)



def plot_nlsolve(
        res_table:pd.DataFrame,
        show:bool=True,
        save:bool=False,
        fname:str|None=None,
        format:str='pdf'
        ) -> None:

    fig, axes = plt.subplots(2, 2, figsize=(10, 10))
    fig:mpl_fig.Figure = fig

    fig.tight_layout(pad=3)

    fig.suptitle('non linear iteration progress', fontsize=13)

    def plot_cols(
            ax:mpl_axes.Axes,
            cols:list,
            ylab:str
            ) -> None:

        for col in cols:
            ax.plot(res_table.index, res_table[col], label=col)
            ax.set_xlabel('iteration nr')
            ax.set_ylabel(ylab)
            ax.legend(frameon=False)


    def plot_xy_s(ax: mpl_axes.Axes, xCol: str, yCol: str) -> None:
        x = res_table[xCol]
        y = res_table[yCol]

        ax.plot(x, y, label='path of stiffness center', color='gray', zorder=1)

        ax.scatter(x.iloc[0], y.iloc[0], color='red', label='First point', zorder=2)
        ax.scatter(x.iloc[-1], y.iloc[-1], color='blue', label='Last point', zorder=2)

        ax.annotate(f'{res_table.index[0]}', (x.iloc[0], y.iloc[0]),
                    textcoords="offset points", xytext=(5, 5), ha='center', zorder=3)
        ax.annotate(f'{res_table.index[-1]}', (x.iloc[-1], y.iloc[-1]),
                    textcoords="offset points", xytext=(5, 5), ha='center', zorder=3)

        ax.set_xlabel('global $x$')
        ax.set_ylabel('global $y$')
        ax.legend(frameon=False)


    v_cols = [col for col in res_table.columns if 'V' in col]
    m_cols = [col for col in res_table.columns if 'M' in col]
    k_cols = [col for col in res_table.columns if 'EI' in col]

    plot_cols(axes[0, 0], v_cols, 'shear forces  $V$')
    plot_cols(axes[0, 1], m_cols, 'moments  $M$')
    plot_cols(axes[1, 0], k_cols, 'bending stiffness  $EI$')
    plot_xy_s(axes[1, 1], 'x_s', 'y_s')

    if save:
        current_time = datetime.now()
        formatted_time = f'{current_time:%Y-%m-%d_%H%M%S}'
        if fname:
            plt.savefig(f'{fname}.{format}')
        else:
            plt.savefig(f'{formatted_time}_nlsolve.{format}')

    if show:
        plt.show()
